@@ -4,7 +4,10 @@ import uuid
 import time
 import pickle
 import collections
+import math
 
+from threading import Thread
+from time import sleep
 from dockets import errors, _global_event_handler_classes, _global_retry_error_classes
 from dockets.pipeline import PipelineObject
 from dockets.metadata import WorkerMetadataRecorder
@@ -46,9 +49,9 @@ class Queue(PipelineObject):
         self.worker_id = uuid.uuid1()
 
         self._wait_time = kwargs.get('wait_time') or 10
-        self._activity_timeout = kwargs.get('timeout', 60)
         self._serializer = kwargs.get('serializer', JsonSerializer())
         self._max_attempts = kwargs.get('max_attempts', 3)
+        self._heartbeat_interval = kwargs.get('heartbeat_interval', 5)
 
         # register event handlers
         self._event_registrar = QueueEventRegistrar(self)
@@ -93,7 +96,6 @@ class Queue(PipelineObject):
     @PipelineObject.with_pipeline
     def pop(self, pipeline):
         pop_pipeline = self.redis.pipeline()
-        self._record_worker_activity(pipeline=pop_pipeline)
         args = [self._queue_key(), self._working_queue_key(), self._wait_time]
         pop_pipeline.execute()
         serialized_envelope = self.redis.brpoplpush(*args)
@@ -157,12 +159,54 @@ class Queue(PipelineObject):
         """
         pipeline.lrem(self._working_queue_key(), serialized_envelope)
 
+    @PipelineObject.with_pipeline
+    def _heartbeat(self, pipeline):
+        """
+        Pushes out this worker's timeout.
+        """
+        pipeline.sadd(self._workers_set_key(), self.worker_id)
+        pipeline.setex(self._worker_activity_key(), 1,
+                       int(math.ceil(self._heartbeat_interval * 5)))
+
+
+    def _start_heartbeat_thread(self, should_stop):
+        """
+        Performs a "heartbeat" every interval. In practice this just pushes out
+        the expiry on this worker's activity key to ensure its items are not
+        reclaimed.
+        """
+        def run_heartbeat():
+            while True:
+                sleep(self._heartbeat_interval)
+                if should_stop():
+                    break
+                self._heartbeat()
+
+        # Do an initial heartbeat to avoid any race conditions. Subsequent
+        # heartbeats will happen in the background thread.
+        self._heartbeat()
+
+        thread = Thread(target=run_heartbeat)
+        thread.daemon = True   # Make sure this doesn't keep the process hanging around if the main thread dies.
+        thread.start()
+        return thread
+
     def run(self, should_continue=None):
         self.register_worker()
         should_continue = should_continue or (lambda: True)
+
+        stopping = False
+
+        heartbeat_thread = self._start_heartbeat_thread(lambda: stopping)
+
         while True:
             if not should_continue():
+                stopping = True
+
+            if stopping:
+                heartbeat_thread.join(self._heartbeat_interval * 2)
                 break
+
             self.run_once()
 
 
@@ -300,12 +344,6 @@ class Queue(PipelineObject):
 
     ## worker handling
 
-    @PipelineObject.with_pipeline
-    def _record_worker_activity(self, pipeline):
-        pipeline.sadd(self._workers_set_key(), self.worker_id)
-        pipeline.setex(self._worker_activity_key(), 1,
-                       self._activity_timeout)
-
     # TODO rewrite using lua scripting
     def _reclaim(self):
         workers = self.redis.smembers(self._workers_set_key())
@@ -336,8 +374,3 @@ SERIALIZATION = Queue.SERIALIZATION
 DESERIALIZATION = Queue.DESERIALIZATION
 
 
-
-class TestQueue(Queue):
-    def process_item(self, item):
-        time.sleep(2)
-        logging.info('in process_item, processing {0}'.format(item))
